"""Constants for formatters."""

WWW_BASE_URL = "https://www.inaturalist.org"

ICONS = {
    "research": "\N{WHITE HEAVY CHECK MARK}",
    "needs_id": "\N{LARGE ORANGE DIAMOND}",
    "casual": "\N{MEDIUM WHITE CIRCLE}",
    "fave": "\N{WHITE MEDIUM STAR}",
    "comment": "\N{LEFT SPEECH BUBBLE}",
    "community": "\N{BUSTS IN SILHOUETTE}",
    "image": "\N{CAMERA}",
    "sound": "\N{SPEAKER WITH ONE SOUND WAVE}",
    "ident": "\N{LABEL}",
}
